
#include <stdio.h>
#include <stdlib.h>
#include "facet.h"

#define UChar	unsigned char
#define UShort	unsigned short

extern void	Error(char *msg);

static void	Set(Facet *facet,double nx,double ny,double nz,
				 double x1,double y1,double z1,
				 double x2,double y2,double z2,
				 double x3,double y3,double z3,UShort flag)
{
	facet->x1=x1; facet->y1=y1; facet->z1=z1;
	if (nx*(y1*z2+y2*z3+y3*z1-z1*y2-z2*y3-z3*y1)+
	    ny*(z1*x2+z2*x3+z3*x1-x1*z2-x2*z3-x3*z1)+
	    nz*(x1*y2+x2*y3+x3*y1-y1*x2-y2*x3-y3*x1)<0.0) {
	    facet->x2=x3; facet->y2=y3; facet->z2=z3;
	    facet->x3=x2; facet->y3=y2; facet->z3=z2;
	}
	else {
	    facet->x2=x2; facet->y2=y2; facet->z2=z2;
	    facet->x3=x3; facet->y3=y3; facet->z3=z3;
	}
#ifdef	FLAG
	facet->FLAG=flag;
#endif
}

static void	Normal(Facet *facet,float *xyz,UShort flag)
{
	Set(facet,xyz[0],xyz[ 1],xyz[ 2],
		  xyz[3],xyz[ 4],xyz[ 5],
		  xyz[6],xyz[ 7],xyz[ 8],
		  xyz[9],xyz[10],xyz[11],flag);
}

static void	Reverse(Facet *facet,float *xyz,UShort flag)
{
	int	i;
	UChar	*j,k;

	for (i=0; i<12; i++) {
	    j=(UChar *)(xyz+i); k=j[0]; j[0]=j[3]; j[3]=k;
				k=j[1]; j[1]=j[2]; j[2]=k;
	}
	j=(UChar *)&flag; k=j[0]; j[0]=j[1]; j[1]=k;

	Normal(facet,xyz,flag);
}

static FILE	*file;
static UChar	S[96];
static size_t	N;
static char	*fmae="facet memory allocation error.";

static Facet	*Binary()
{
	Facet	*facet=NULL;
	float	*xyz=(float *)S,
		*nxyz=(float *)(S+84);
	UShort	flag;
	char	*fdre="binary facet data read error.";
	size_t	n;

	UShort	us=255;
	UChar	*uc=(UChar *)&us;
	void	(*set)()=(void (*)())((*uc==us)?Normal:Reverse);

	if ((N=(size_t)S[80]|((size_t)S[81]<< 8)|
			     ((size_t)S[82]<<16)|((size_t)S[83]<<24))>0) {
	    if ((facet=(Facet *)malloc(sizeof(Facet)*N))==NULL) Error(fmae);

	    if (fread(xyz+3,sizeof(float),9,file)!=9 ||
		fread(&flag,sizeof(UShort),1,file)!=1) Error(fdre);

	    xyz[0]=nxyz[0]; xyz[1]=nxyz[1]; xyz[2]=nxyz[2];
	    (*set)(facet,xyz,flag);

	    for (n=1; n<N; n++) {
		if (fread(xyz,sizeof(float),12,file)!=12 ||
		    fread(&flag,sizeof(UShort),1,file)!=1) Error(fdre);

		(*set)(facet+n,xyz,flag);
	    }
	}
	return facet;
}

static int	I,C;

static int	GetC()
{
	return C=(I<96)?(int)S[I++]:fgetc(file);
}

static int	GetS()
{
	UChar	*s=S;

	while (GetC()==' ' || C=='\r' || C=='\n' || C=='\t') ;

	if (C==EOF) return 0;

	do {
	    *s=(UChar)((C<'A' || C>'Z')?C:C|32); if (++s==S+96) return 0;
	} while (GetC()!=' ' && C!='\r' && C!='\n' && C!='\t' && C!=EOF);

	*s='\0'; return 1;
}

static int	Compare(char *KW)
{
	char	*s=(char *)S;

	while (*s!='\0' || *KW!='\0') if (*(s++)!=(*(KW++))) return 0;

	return 1;
}

static Facet	*ASCII()
{
	char	*fdre="ASCII facet data read error.";
	double	nx,ny,nz,x1,y1,z1,x2,y2,z2,x3,y3,z3;
	size_t	n;
	Facet	*facet=NULL;

	I=0;
	if (!GetS() || !Compare("solid")) Error(fdre);

	while (C!='\n') if (GetC()==EOF) Error(fdre);

	N=0;
	for (;;) {
	    if (!GetS()) Error(fdre);

	    if (Compare("endsolid")) break;

	    if (!Compare("facet") || !GetS() ||
		!Compare("normal") || !GetS() ||
		sscanf((char *)S,"%lf",&nx)!=1 || !GetS() ||
		sscanf((char *)S,"%lf",&ny)!=1 || !GetS() ||
		sscanf((char *)S,"%lf",&nz)!=1 || !GetS() ||
		!Compare("outer") || !GetS() ||
		!Compare("loop") || !GetS() ||
		!Compare("vertex") || !GetS() ||
		sscanf((char *)S,"%lf",&x1)!=1 || !GetS() ||
		sscanf((char *)S,"%lf",&y1)!=1 || !GetS() ||
		sscanf((char *)S,"%lf",&z1)!=1 || !GetS() ||
		!Compare("vertex") || !GetS() ||
		sscanf((char *)S,"%lf",&x2)!=1 || !GetS() ||
		sscanf((char *)S,"%lf",&y2)!=1 || !GetS() ||
		sscanf((char *)S,"%lf",&z2)!=1 || !GetS() ||
		!Compare("vertex") || !GetS() ||
		sscanf((char *)S,"%lf",&x3)!=1 || !GetS() ||
		sscanf((char *)S,"%lf",&y3)!=1 || !GetS() ||
		sscanf((char *)S,"%lf",&z3)!=1 || !GetS() ||
		!Compare("endloop") || !GetS() ||
		!Compare("endfacet")) Error(fdre);

	    n=N++;
	    if ((facet=(Facet *)realloc(facet,
					sizeof(Facet)*N))==NULL) Error(fmae);

	    Set(facet+n,nx,ny,nz,x1,y1,z1,x2,y2,z2,x3,y3,z3,0);
	}
	while (C!='\n' && GetC()!=EOF) ;

	return facet;
}

void	ReadFacet(char *path,size_t *n,Facet **facet)
{
	if (*path=='-')
	    file=stdin;
	else
	    if ((file=fopen(path,"rb"))==NULL) Error("STL file not found.");

	if (fread(S,sizeof(UChar),96,file)!=96) Error("facet data read error.");

	for (I=0; I<96; I++)
	    if (S[I]!='\t' &&
		S[I]!='\n' &&
		S[I]!='\r' && (S[I]<' ' || S[I]>'~')) break;

	*facet=(I<96)?Binary():ASCII();

	if ((*n=N)==0) Error("no facet.");

	if (file!=stdin) (void)fclose(file);
}
