
#if	!defined(SI_P) && !defined(SI_G)
#error macro SI_P and SI_G are not defined.
#endif

#ifdef	DOUBLE
#define FLOAT	double
#define LOG	log
#define SQRT	sqrt
#define RSQRT	rsqrt
#define ATAN2	atan2
#define EPS	2.22045e-16
#else
#define FLOAT	float
#define LOG	logf
#define SQRT	sqrtf
#define RSQRT	rsqrtf
#define ATAN2	atan2f
#define EPS	(float)1.19209e-7
#endif

#define SIZE_T	unsigned

#ifdef	KEPLER

typedef struct {
		FLOAT	x1,y1,z1,x2,y2,z2,x3,y3,z3,im;
	} SI;

__device__ inline FLOAT	LI(FLOAT a,FLOAT b,FLOAT c)
{
	FLOAT	d=a*b;
#ifdef	N_O_E
	return LOG((d+(FLOAT)1.0)/(d-(FLOAT)1.0))*a*c;
#else
	FLOAT	r=(d-(FLOAT)1.0)/(d+(FLOAT)1.0);

	return (r<EPS)?(FLOAT)0.0:-LOG(r)*a*c;
#endif
}

__global__ void	Kernel(SIZE_T No,SIZE_T Nt,
		       FLOAT *X0,FLOAT *Y0,FLOAT *Z0,
		       FLOAT *X1,FLOAT *Y1,FLOAT *Z1,
		       FLOAT *X2,FLOAT *Y2,FLOAT *Z2,
		       FLOAT *X3,FLOAT *Y3,FLOAT *Z3,FLOAT *IM
#ifdef	SI_P
		      ,FLOAT *P
#endif
#ifdef	SI_G
		      ,FLOAT *Gx,FLOAT *Gy,FLOAT *Gz
#endif
		      )
{
	SIZE_T	o=threadIdx.x+blockDim.x*(blockIdx.x+gridDim.x*blockIdx.y);

	if (o<No)
{
	FLOAT			x0=X0[o],y0=Y0[o],z0=Z0[o];
	SIZE_T			t,n;

	extern __shared__ SI	si[/* blockDim.x */];
	SI			*SI_T=si+threadIdx.x;
#ifdef	SI_P
	FLOAT			p=(FLOAT)0.0;
#endif
#ifdef	SI_G
	FLOAT			gx=(FLOAT)0.0,gy=(FLOAT)0.0,gz=(FLOAT)0.0;
#endif
	for (t=threadIdx.x; t<Nt; t+=blockDim.x)
{
	__syncthreads();
	SI_T->x1=X1[t]; SI_T->y1=Y1[t]; SI_T->z1=Z1[t];
	SI_T->x2=X2[t]; SI_T->y2=Y2[t]; SI_T->z2=Z2[t];
	SI_T->x3=X3[t]; SI_T->y3=Y3[t]; SI_T->z3=Z3[t]; SI_T->im=IM[t];
	__syncthreads();

	for (n=0; n<blockDim.x; n++)
{
	SI	*si_n=si+n;
	FLOAT	x1=si_n->x1-x0,y1=si_n->y1-y0,z1=si_n->z1-z0,
		x2=si_n->x2-x0,y2=si_n->y2-y0,z2=si_n->z2-z0,
		x3=si_n->x3-x0,y3=si_n->y3-y0,z3=si_n->z3-z0,im=si_n->im,
		d1=SQRT(x1*x1+y1*y1+z1*z1),
		d2=SQRT(x2*x2+y2*y2+z2*z2),
		d3=SQRT(x3*x3+y3*y3+z3*z3),
		x12=x2-x1,y12=y2-y1,z12=z2-z1,
		x23=x3-x2,y23=y3-y2,z23=z3-z2,
		x31=x1-x3,y31=y1-y3,z31=z1-z3,
		nx=im*(z12*y31-y12*z31),
		ny=im*(x12*z31-z12*x31),
		nz=im*(y12*x31-x12*y31),
		f=x1*nx+y1*ny+z1*nz,
		g=LI(RSQRT(x12*x12+y12*y12+z12*z12),d1+d2,
		     x1*(y12*nz-z12*ny)+y1*(z12*nx-x12*nz)+z1*(x12*ny-y12*nx))
		 +LI(RSQRT(x23*x23+y23*y23+z23*z23),d2+d3,
		     x2*(y23*nz-z23*ny)+y2*(z23*nx-x23*nz)+z2*(x23*ny-y23*nx))
		 +LI(RSQRT(x31*x31+y31*y31+z31*z31),d3+d1,
		     x3*(y31*nz-z31*ny)+y3*(z31*nx-x31*nz)+z3*(x31*ny-y31*nx))
		 -f*(FLOAT)2.0*ATAN2(x1*y2*z3+x2*y3*z1+x3*y1*z2
				    -x3*y2*z1-x1*y3*z2-x2*y1*z3,
				     d1*d2*d3+d1*(x2*x3+y2*y3+z2*z3)
					     +d2*(x3*x1+y3*y1+z3*z1)
					     +d3*(x1*x2+y1*y2+z1*z2));
#ifdef	SI_P
	p+=(f*g);
#endif
#ifdef	SI_G
	gx-=(nx*g); gy-=(ny*g); gz-=(nz*g);
#endif
}}
	__syncthreads();
#ifdef  SI_P
	P[o]=p*(FLOAT)0.5;
#endif
#ifdef	SI_G
	Gx[o]=gx; Gy[o]=gy; Gz[o]=gz;
#endif
}}

#else	/* KEPLER */

typedef struct {
		FLOAT	x1,y1,z1,x2,y2,z2,x3,y3,z3,
			nx,ny,nz,id12,ox12,oy12,oz12,
				 id23,ox23,oy23,oz23,
				 id31,ox31,oy31,oz31;
	} SI;

__device__ inline FLOAT	LI(FLOAT d,FLOAT e)
{
#ifdef	N_O_E
	return LOG((d+(FLOAT)1.0)/(d-(FLOAT)1.0))*e;
#else
	FLOAT	r=(d-(FLOAT)1.0)/(d+(FLOAT)1.0);

	return (r<EPS)?(FLOAT)0.0:-LOG(r)*e;
#endif
}

__global__ void	Kernel(SIZE_T No,SIZE_T Nt,
		       FLOAT *X0,FLOAT *Y0,FLOAT *Z0,
		       FLOAT *X1,FLOAT *Y1,FLOAT *Z1,
		       FLOAT *X2,FLOAT *Y2,FLOAT *Z2,
		       FLOAT *X3,FLOAT *Y3,FLOAT *Z3,FLOAT *IM
#ifdef	SI_P
		      ,FLOAT *P
#endif
#ifdef	SI_G
		      ,FLOAT *Gx,FLOAT *Gy,FLOAT *Gz
#endif
		      )
{
	SIZE_T	o=threadIdx.x+blockDim.x*(blockIdx.x+gridDim.x*blockIdx.y);

	if (o<No)
{
	FLOAT			x0=X0[o],y0=Y0[o],z0=Z0[o];
	SIZE_T			t,n;

	extern __shared__ SI	si[/* blockDim.x */];
	SI			*SI_T=si+threadIdx.x;
#ifdef	SI_P
	FLOAT			p=(FLOAT)0.0;
#endif
#ifdef	SI_G
	FLOAT			gx=(FLOAT)0.0,gy=(FLOAT)0.0,gz=(FLOAT)0.0;
#endif
	for (t=threadIdx.x; t<Nt; t+=blockDim.x)
{
	__syncthreads();
{
	FLOAT	x1=X1[t],y1=Y1[t],z1=Z1[t],
		x2=X2[t],y2=Y2[t],z2=Z2[t],
		x3=X3[t],y3=Y3[t],z3=Z3[t],im=IM[t],
		x12=x2-x1,y12=y2-y1,z12=z2-z1,
		x23=x3-x2,y23=y3-y2,z23=z3-z2,
		x31=x1-x3,y31=y1-y3,z31=z1-z3,
		nx=im*(z12*y31-y12*z31),
		ny=im*(x12*z31-z12*x31),
		nz=im*(y12*x31-x12*y31),
		id12=RSQRT(x12*x12+y12*y12+z12*z12),
		id23=RSQRT(x23*x23+y23*y23+z23*z23),
		id31=RSQRT(x31*x31+y31*y31+z31*z31);

	SI_T->x1=x1; SI_T->y1=y1; SI_T->z1=z1;
	SI_T->x2=x2; SI_T->y2=y2; SI_T->z2=z2;
	SI_T->x3=x3; SI_T->y3=y3; SI_T->z3=z3;
	SI_T->nx=nx; SI_T->ny=ny; SI_T->nz=nz;
	SI_T->id12=id12; SI_T->ox12=id12*(y12*nz-z12*ny);
			 SI_T->oy12=id12*(z12*nx-x12*nz);
			 SI_T->oz12=id12*(x12*ny-y12*nx);
	SI_T->id23=id23; SI_T->ox23=id23*(y23*nz-z23*ny);
			 SI_T->oy23=id23*(z23*nx-x23*nz);
			 SI_T->oz23=id23*(x23*ny-y23*nx);
	SI_T->id31=id31; SI_T->ox31=id31*(y31*nz-z31*ny);
			 SI_T->oy31=id31*(z31*nx-x31*nz);
			 SI_T->oz31=id31*(x31*ny-y31*nx);
}
	__syncthreads();

	for (n=0; n<blockDim.x; n++)
{
	SI	*si_n=si+n;
	FLOAT	x1=si_n->x1-x0,y1=si_n->y1-y0,z1=si_n->z1-z0,
		x2=si_n->x2-x0,y2=si_n->y2-y0,z2=si_n->z2-z0,
		x3=si_n->x3-x0,y3=si_n->y3-y0,z3=si_n->z3-z0,
		d1=SQRT(x1*x1+y1*y1+z1*z1),
		d2=SQRT(x2*x2+y2*y2+z2*z2),
		d3=SQRT(x3*x3+y3*y3+z3*z3),
		f=x1*si_n->nx+y1*si_n->ny+z1*si_n->nz,
		g=LI(si_n->id12*(d1+d2),
		     x1*si_n->ox12+y1*si_n->oy12+z1*si_n->oz12)
		 +LI(si_n->id23*(d2+d3),
		     x2*si_n->ox23+y2*si_n->oy23+z2*si_n->oz23)
		 +LI(si_n->id31*(d3+d1),
		     x3*si_n->ox31+y3*si_n->oy31+z3*si_n->oz31)
		 -f*(FLOAT)2.0*ATAN2(x1*y2*z3+x2*y3*z1+x3*y1*z2
				    -x3*y2*z1-x1*y3*z2-x2*y1*z3,
				     d1*d2*d3+d1*(x2*x3+y2*y3+z2*z3)
					     +d2*(x3*x1+y3*y1+z3*z1)
					     +d3*(x1*x2+y1*y2+z1*z2));
#ifdef	SI_P
	p+=(f*g);
#endif
#ifdef	SI_G
	gx-=(si_n->nx*g); gy-=(si_n->ny*g); gz-=(si_n->nz*g);
#endif
}}
	__syncthreads();
#ifdef  SI_P
	P[o]=p*(FLOAT)0.5;
#endif
#ifdef	SI_G
	Gx[o]=gx; Gy[o]=gy; Gz[o]=gz;
#endif
}}

#endif	/* KEPLER */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

extern "C" void	Error(char *msg);

#define CALL(name,arg)	\
	if (name arg !=cudaSuccess) Error("error in " #name "().")

extern "C" void	SurfaceIntegral(size_t No,size_t Nt,
				double *X0,double *Y0,double *Z0,
				double *X1,double *Y1,double *Z1,
				double *X2,double *Y2,double *Z2,
				double *X3,double *Y3,double *Z3
#ifdef	SI_P
			       ,double *P
#endif
#ifdef	SI_G
			       ,double *Gx,double *Gy,double *Gz
#endif
			       )
{
	SIZE_T	K,L,M;
	size_t	no,nt,n;
	FLOAT	*W,*x0,*y0,*z0,*x1,*y1,*z1,*x2,*y2,*z2,*x3,*y3,*z3,*im;
#ifdef	SI_P
	FLOAT	*p;
#endif
#ifdef	SI_G
	FLOAT	*gx,*gy,*gz;
#endif
{
	char			*ev=getenv("CUDA_GPU"),
				*nd="no device assigned by CUDA_GPU.";
	int			dc,
				cg=0;
	cudaDeviceProp		dp;
	cudaFuncAttributes	fa;

	if (ev!=NULL && (sscanf(ev,"%d",&cg)!=1 || cg<0)) Error(nd);

	CALL(cudaGetDeviceCount,(&dc)); if (cg>=dc) Error(nd);

	CALL(cudaGetDeviceProperties,(&dp,cg));
	if (dp.major<1) Error("device not supported by CUDA.");

	CALL(cudaSetDevice,(cg));

	CALL(cudaFuncGetAttributes,(&fa,Kernel));

	K=(SIZE_T)((dp.sharedMemPerBlock-fa.sharedSizeBytes)/sizeof(SI));
	if (K*fa.numRegs>dp.regsPerBlock) K=dp.regsPerBlock/fa.numRegs;
	if (K>dp.maxThreadsPerBlock) K=dp.maxThreadsPerBlock;
	if (K>fa.maxThreadsPerBlock) K=fa.maxThreadsPerBlock;
	K=(K/dp.warpSize)*dp.warpSize;

#ifdef	DEBUG
	(void)fprintf(stderr,"%u\t%u\t%u\t%d\t%d\t%d\t%d\t%d\t%u\n",
			     (SIZE_T)dp.sharedMemPerBlock,
			     (SIZE_T)fa.sharedSizeBytes,
			     (SIZE_T)sizeof(SI),
			     dp.regsPerBlock,
			     fa.numRegs,
			     dp.maxThreadsPerBlock,
			     fa.maxThreadsPerBlock,
			     dp.warpSize,
			     K);
#endif
	L=dp.maxGridSize[0];
}
	no=(size_t)K*(M=(SIZE_T)((No+K-1)/K));
	nt=(size_t)K*   (SIZE_T)((Nt+K-1)/K) ;

#define SOF		sizeof(FLOAT)
#define CMA(N,D)	CALL(cudaMalloc,((void **)&D,SOF*N))
#define CMC(N,D)	CALL(cudaMemcpy,(D,W,SOF*N,cudaMemcpyHostToDevice))
#define H2D(OT,ot,D,H)	for (n=0; n<OT; n++) W[n]=H[n];\
			for (; n<ot; n++) W[n]=H[0];\
			CMA(ot,D);\
			CMC(ot,D)

	n=(no>nt)?no:nt; CALL(cudaMallocHost,((void **)&W,SOF*n));

	H2D(No,no,x0,X0); H2D(No,no,y0,Y0); H2D(No,no,z0,Z0);
	H2D(Nt,nt,x1,X1); H2D(Nt,nt,y1,Y1); H2D(Nt,nt,z1,Z1);
	H2D(Nt,nt,x2,X2); H2D(Nt,nt,y2,Y2); H2D(Nt,nt,z2,Z2);
	H2D(Nt,nt,x3,X3); H2D(Nt,nt,y3,Y3); H2D(Nt,nt,z3,Z3);

	for (n=0; n<Nt; n++)
{
	double	xa=X2[n]-X1[n],ya=Y2[n]-Y1[n],za=Z2[n]-Z1[n],
		xb=X3[n]-X1[n],yb=Y3[n]-Y1[n],zb=Z3[n]-Z1[n],
		xc=ya*zb-za*yb,yc=za*xb-xa*zb,zc=xa*yb-ya*xb;

	W[n]=1.0/sqrt(xc*xc+yc*yc+zc*zc);
}
	for (; n<nt; n++) W[n]=0.0;
	CMA(nt,im);
	CMC(nt,im);

#ifdef	SI_P
	CMA(no,p);
#endif
#ifdef	SI_G
	CMA(no,gx); CMA(no,gy); CMA(no,gz);
#endif
{
	SIZE_T	X,Y;

	if (M<=L) {
	    X=M; Y=1;
	}
	else
{
	SIZE_T	Z,x,y,z;

	if ((Z=(Y=(M+L-1)/L)*(X=L))!=M)
	    for (x=L-1; x>=(y=(M+x-1)/x); x--) 
		if ((z=y*x)<Z) {
		    X=x; Y=y; if ((Z=z)==M) break;
		}
}{
	dim3	g(X,Y);

	Kernel<<<g,K,sizeof(SI)*K>>>((SIZE_T)no,(SIZE_T)nt,
				     x0,y0,z0,x1,y1,z1,x2,y2,z2,x3,y3,z3,im
#ifdef	SI_P
				    ,p
#endif
#ifdef	SI_G
				    ,gx,gy,gz
#endif
				    );
}}
	if (cudaGetLastError()!=cudaSuccess) Error("error in Kernel().");

#define CF(D)		CALL(cudaFree,(D))
#define D2H(H,D)	CALL(cudaMemcpy,(W,D,SOF*No,cudaMemcpyDeviceToHost));\
			CF(D);\
			for (n=0; n<No; n++) H[n]=W[n]

#ifdef	SI_G
	D2H(Gz,gz); D2H(Gy,gy); D2H(Gx,gx);
#endif
#ifdef	SI_P
	D2H(P,p);
#endif
	CF(im); CF(z3); CF(y3); CF(x3);
		CF(z2); CF(y2); CF(x2);
		CF(z1); CF(y1); CF(x1);
		CF(z0); CF(y0); CF(x0); CALL(cudaFreeHost,(W));
}
