
#include <stdio.h>
#include <stdlib.h>
#include "facet.h"

extern void	Error(char *msg),
		SurfaceIntegral(size_t No,size_t Nt,
				double *X0,double *Y0,double *Z0,
				double *X1,double *Y1,double *Z1,
				double *X2,double *Y2,double *Z2,
				double *X3,double *Y3,double *Z3,
				double *P,double *Gx,double *Gy,double *Gz);

#define LEN	1024

int	main(int argc,char **argv)
{
	size_t	Nt,n,No;
	Facet	*facet;
	double	*X1,*Y1,*Z1,*X2,*Y2,*Z2,*X3,*Y3,*Z3,V,X,Y,Z,x,y,z,v,
		*X0,*Y0,*Z0,*P,*Gx,*Gy,*Gz;
	FILE	*file;
	char	text[LEN],
		*mae="memory allocation error.";

	if (argc!=2 && argc!=3) Error("usage : stl_pxyz STL {point}");

	ReadFacet(argv[1],&Nt,&facet);

#define MA(A,N)	(A=(double *)malloc(sizeof(double)*(N)))==NULL

	if (MA(X1,Nt) || MA(Y1,Nt) || MA(Z1,Nt) ||
	    MA(X2,Nt) || MA(Y2,Nt) || MA(Z2,Nt) ||
	    MA(X3,Nt) || MA(Y3,Nt) || MA(Z3,Nt)) Error(mae);

	V=X=Y=Z=0.0;
	for (n=0; n<Nt; n++) {
	    x=((X1[n]=facet[n].x1)+(X2[n]=facet[n].x2)+(X3[n]=facet[n].x3));
            y=((Y1[n]=facet[n].y1)+(Y2[n]=facet[n].y2)+(Y3[n]=facet[n].y3));
            z=((Z1[n]=facet[n].z1)+(Z2[n]=facet[n].z2)+(Z3[n]=facet[n].z3));
            V+=(v=X1[n]*Y2[n]*Z3[n]+X2[n]*Y3[n]*Z1[n]+X3[n]*Y1[n]*Z2[n]
                 -X3[n]*Y2[n]*Z1[n]-X1[n]*Y3[n]*Z2[n]-X2[n]*Y1[n]*Z3[n]);
            X+=(v*x); Y+=(v*y); Z+=(v*z);
	}
	(void)printf("%le\t%le\t%le\t%le\n",V/6.0,X/V/4.0,Y/V/4.0,Z/V/4.0);

	free(facet);

	if (argc==2) {
	    if (MA(X0,Nt) || MA(Y0,Nt) || MA(Z0,Nt) || MA(P,Nt)) Error(mae);

	    for (n=0; n<Nt; n++) {
		X0[n]=(X1[n]+X2[n]+X3[n])/3.0;
		Y0[n]=(Y1[n]+Y2[n]+Y3[n])/3.0;
		Z0[n]=(Z1[n]+Z2[n]+Z3[n])/3.0;
	    }
	    SurfaceIntegral(Nt,Nt,
			    X0,Y0,Z0,X1,Y1,Z1,X2,Y2,Z2,X3,Y3,Z3,P,X0,Y0,Z0);

	    for (n=0; n<Nt; n++)
		(void)printf("%le\t%le\t%le\t"
			     "%le\t%le\t%le\t"
			     "%le\t%le\t%le\t%le\t%le\t%le\t%le\n",
			     X1[n],Y1[n],Z1[n],
			     X2[n],Y2[n],Z2[n],
			     X3[n],Y3[n],Z3[n],P[n],X0[n],Y0[n],Z0[n]);
	}
	else {
	    if (argv[2][0]=='-')
		file=stdin;
	    else
		if ((file=fopen(argv[2],"r"))==NULL)
		    Error("point file not found.");
 
	    No=0; X0=Y0=Z0=NULL;
	    while (fgets(text,LEN,file)!=NULL &&
		   sscanf(text,"%lf %lf %lf",&x,&y,&z)==3) {
		n=No++;

#define RA(A,N)	(A=(double *)realloc(A,sizeof(double)*(N)))==NULL

		if (RA(X0,No) || RA(Y0,No) || RA(Z0,No))
		    Error("memory re-allocation error.");

		X0[n]=x; Y0[n]=y; Z0[n]=z;
	    }
	    if (file!=stdin) fclose(file);

	    if (No!=0) {
		if (MA(P,No) || MA(Gx,No) || MA(Gy,No) || MA(Gz,No)) Error(mae);

		SurfaceIntegral(No,Nt,
				X0,Y0,Z0,X1,Y1,Z1,X2,Y2,Z2,X3,Y3,Z3,P,Gx,Gy,Gz);

		for (n=0; n<No; n++)
		    (void)printf("%le\t%le\t%le\t%le\t%le\t%le\t%le\n",
				 X0[n],Y0[n],Z0[n],P[n],Gx[n],Gy[n],Gz[n]);
	    }
	}
	return 0;
}
