
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "ph.h"

extern void	Error(char *msg),
		SurfaceIntegral(double x0,double y0,double z0,
				double x1,double y1,double z1,
				double x2,double y2,double z2,
				double x3,double y3,double z3,
				double *P,double *X,double *Y,double *Z);

static double	V,X,Y,Z;

static void	SI(double x1,double y1,double z1,
		   double x2,double y2,double z2,
		   double x3,double y3,double z3)
{
	double	v=x1*y2*z3+x2*y3*z1+x3*y1*z2-x3*y2*z1-x1*y3*z2-x2*y1*z3;

	V+=v; X+=(v*(x1+x2+x3)); Y+=(v*(y1+y2+y3)); Z+=(v*(z1+z2+z3));
}

static void	Unit(double x,double y,double z,
		     double *px,double *py,double *pz)
{
	double	l=sqrt(x*x+y*y+z*z);

	*px=x/l; *py=y/l; *pz=z/l;
}

#define FUNCTION(NAME,CALL)	\
static void	NAME(int level,double x1,double y1,double z1,\
			       double x2,double y2,double z2,\
			       double x3,double y3,double z3)\
{\
	double	xa,ya,za,xb,yb,zb,xc,yc,zc;\
\
	if (level==0)\
	    CALL;\
	else {\
	    Unit((x1+x2)/2.0,(y1+y2)/2.0,(z1+z2)/2.0,&xa,&ya,&za);\
	    Unit((x2+x3)/2.0,(y2+y3)/2.0,(z2+z3)/2.0,&xb,&yb,&zb);\
	    Unit((x3+x1)/2.0,(y3+y1)/2.0,(z3+z1)/2.0,&xc,&yc,&zc);\
	    NAME(--level,xa,ya,za,xb,yb,zb,xc,yc,zc);\
	    NAME(  level,x1,y1,z1,xa,ya,za,xc,yc,zc);\
	    NAME(  level,x2,y2,z2,xb,yb,zb,xa,ya,za);\
	    NAME(  level,x3,y3,z3,xc,yc,zc,xb,yb,zb);\
	}\
}

FUNCTION(VXYZ,SI(x1,y1,z1,x2,y2,z2,x3,y3,z3))

static double	x,y,z,P;

FUNCTION(PXYZ,SurfaceIntegral(x,y,z,x1,y1,z1,x2,y2,z2,x3,y3,z3,&P,&X,&Y,&Z))

int	main(int argc,char **argv)
{
	Trigon	*T;
	int	n,
		level=0,
		N=Polyhedron(&T);

	if (argc!=1 && argc!=2 && argc!=4 && argc!=5)
	    Error("usage : globe {level} {lon lat rot}");

	if (argc==2 || argc==5)
	    if ((level=atoi(argv[1]))<0) Error("bad recursion level");

	if (argc==4 || argc==5)
	    RotatePolyhedron(atof(argv[argc-3]),
			     atof(argv[argc-2]),
			     atof(argv[argc-1]));

	V=X=Y=Z=0.0;
	for (n=0; n<N; n++)
	    VXYZ(level,T[n].a->x,T[n].a->y,T[n].a->z,
		       T[n].b->x,T[n].b->y,T[n].b->z,
		       T[n].c->x,T[n].c->y,T[n].c->z);

	(void)printf("%lf\t%lf\t%lf\t%lf\n",V/6.0,X/V/4.0,Y/V/4.0,Z/V/4.0);

	while (scanf("%lf %lf %lf",&x,&y,&z)==3) {
	    P=X=Y=Z=0.0;
	    for (n=0; n<N; n++)
		PXYZ(level,T[n].a->x,T[n].a->y,T[n].a->z,
			   T[n].b->x,T[n].b->y,T[n].b->z,
			   T[n].c->x,T[n].c->y,T[n].c->z);

	    (void)printf("%lg\t%lg\t%lg\t%lf\t%lf\t%lf\t%lf\n",x,y,z,P,X,Y,Z);
	}
	return 0;
}
