
#include <math.h>

#define D(i,j,k)	sqrt(i*i+j*j+k*k)

#ifdef	N_O_E	/* not on edge */
#define E(e,l)	l
#else
#define E(e,l)	((e<1e-9)?0.0:l)
#endif

void	SurfaceIntegral(double X0,double Y0,double Z0,
			double X1,double Y1,double Z1,
			double X2,double Y2,double Z2,
			double X3,double Y3,double Z3,
			double *P,double *X,double *Y,double *Z) {
	double	x12=X2-X1,y12=Y2-Y1,z12=Z2-Z1,D12=D(x12,y12,z12),
		x23=X3-X2,y23=Y3-Y2,z23=Z3-Z2,D23=D(x23,y23,z23),
		x31=X1-X3,y31=Y1-Y3,z31=Z1-Z3,D31=D(x31,y31,z31),
		a=z12*y31-y12*z31,
		b=x12*z31-z12*x31,
		c=y12*x31-x12*y31,d=D(a,b,c),NX=a/d,NY=b/d,NZ=c/d,
		OX12=y12*NZ-z12*NY,OY12=z12*NX-x12*NZ,OZ12=x12*NY-y12*NX,
		OX23=y23*NZ-z23*NY,OY23=z23*NX-x23*NZ,OZ23=x23*NY-y23*NX,
		OX31=y31*NZ-z31*NY,OY31=z31*NX-x31*NZ,OZ31=x31*NY-y31*NX,

		x1=X1-X0,y1=Y1-Y0,z1=Z1-Z0,d1=D(x1,y1,z1),
		x2=X2-X0,y2=Y2-Y0,z2=Z2-Z0,d2=D(x2,y2,z2),
		x3=X3-X0,y3=Y3-Y0,z3=Z3-Z0,d3=D(x3,y3,z3),
		e12=d1+d2-D12,e23=d2+d3-D23,e31=d3+d1-D31,
/*
		f=((x1+x2+x3)*NX+(y1+y2+y3)*NY+(z1+z2+z3)*NZ)/3.0,
*/
		f=x1*NX+y1*NY+z1*NZ,
		g=E(e12,log((d1+d2+D12)/e12)*(x1*OX12+y1*OY12+z1*OZ12)/D12)
		 +E(e23,log((d2+d3+D23)/e23)*(x2*OX23+y2*OY23+z2*OZ23)/D23)
		 +E(e31,log((d3+d1+D31)/e31)*(x3*OX31+y3*OY31+z3*OZ31)/D31)
		 -f*2.0*atan2(x1*y2*z3+x2*y3*z1+x3*y1*z2
			     -x3*y2*z1-x1*y3*z2-x2*y1*z3,
			      d1*d2*d3+d1*(x2*x3+y2*y3+z2*z3)
				      +d2*(x3*x1+y3*y1+z3*z1)
				      +d3*(x1*x2+y1*y2+z1*z2));

	*P+=(f*g/2.0); *X-=(NX*g); *Y-=(NY*g); *Z-=(NZ*g);
}
