
#if	!defined(SI_P) && !defined(SI_G)
#error macro SI_P and SI_G are not defined.
#endif

#include <stdlib.h>
#include <math.h>
#include "facet.h"

extern void	Error(char *msg);

#define X1(n)	facet[n].x1
#define Y1(n)	facet[n].y1
#define Z1(n)	facet[n].z1
#define X2(n)	facet[n].x2
#define Y2(n)	facet[n].y2
#define Z2(n)	facet[n].z2
#define X3(n)	facet[n].x3
#define Y3(n)	facet[n].y3
#define Z3(n)	facet[n].z3

typedef struct {
		double	nx,ny,nz,id12,ox12,oy12,oz12,
				 id23,ox23,oy23,oz23,
				 id31,ox31,oy31,oz31;
	} SI;

#define D(u,v,w)	sqrt(u*u+v*v+w*w)

static SI	*BeginSI(size_t N,Facet *facet)
{
	SI	*si=(SI *)malloc(sizeof(SI)*N);
	size_t	n;

	if (si==NULL) Error("memory allocation error for SI.");

	for (n=0; n<N; n++) {

double	x1=X1(n),y1=Y1(n),z1=Z1(n),
	x2=X2(n),y2=Y2(n),z2=Z2(n),
	x3=X3(n),y3=Y3(n),z3=Z3(n),
	x12=x2-x1,y12=y2-y1,z12=z2-z1,id12=1.0/D(x12,y12,z12),
	x23=x3-x2,y23=y3-y2,z23=z3-z2,id23=1.0/D(x23,y23,z23),
	x31=x1-x3,y31=y1-y3,z31=z1-z3,id31=1.0/D(x31,y31,z31),
	a=z12*y31-y12*z31,b=x12*z31-z12*x31,c=y12*x31-x12*y31,d=D(a,b,c),
	nx=a/d,ny=b/d,nz=c/d;

	    si[n].nx=nx; si[n].ny=ny; si[n].nz=nz;
	    si[n].id12=id12; si[n].ox12=id12*(y12*nz-z12*ny);
			     si[n].oy12=id12*(z12*nx-x12*nz);
			     si[n].oz12=id12*(x12*ny-y12*nx);
	    si[n].id23=id23; si[n].ox23=id23*(y23*nz-z23*ny);
			     si[n].oy23=id23*(z23*nx-x23*nz);
			     si[n].oz23=id23*(x23*ny-y23*nx);
	    si[n].id31=id31; si[n].ox31=id31*(y31*nz-z31*ny);
			     si[n].oy31=id31*(z31*nx-x31*nz);
			     si[n].oz31=id31*(x31*ny-y31*nx);
	}
	return si;
}

static double	LI(double d,double e)
{
#ifdef	N_O_E
	return log((d+1.0)/(d-1.0))*e;
#else
	double	r=(d-1.0)/(d+1.0);

	return (r<2.22045e-16)?0.0:-log(r)*e;
#endif
}

static void	CalculateSI(size_t N,Facet *facet,SI *si,
			    double x0,double y0,double z0
#ifdef	SI_P
			    ,double *P
#endif
#ifdef	SI_G
			    ,double *Gx,double *Gy,double *Gz
#endif
			   )
{
	size_t	n;
#ifdef	SI_P
	double	p=0.0;
#endif
#ifdef	SI_G
	double	gx=0.0,gy=0.0,gz=0.0;
#endif
	for (n=0; n<N; n++) {

double	x1=X1(n)-x0,y1=Y1(n)-y0,z1=Z1(n)-z0,d1=D(x1,y1,z1),
	x2=X2(n)-x0,y2=Y2(n)-y0,z2=Z2(n)-z0,d2=D(x2,y2,z2),
	x3=X3(n)-x0,y3=Y3(n)-y0,z3=Z3(n)-z0,d3=D(x3,y3,z3),
/*
	f=((x1+x2+x3)*si[n].nx+(y1+y2+y3)*si[n].ny+(z1+z2+z3)*si[n].nz)/3.0,
*/
	f=x1*si[n].nx+y1*si[n].ny+z1*si[n].nz,
	g=LI(si[n].id12*(d1+d2),x1*si[n].ox12+y1*si[n].oy12+z1*si[n].oz12)
	 +LI(si[n].id23*(d2+d3),x2*si[n].ox23+y2*si[n].oy23+z2*si[n].oz23)
	 +LI(si[n].id31*(d3+d1),x3*si[n].ox31+y3*si[n].oy31+z3*si[n].oz31)
	 -f*2.0*atan2(x1*y2*z3+x2*y3*z1+x3*y1*z2
		     -x3*y2*z1-x1*y3*z2-x2*y1*z3,
		      d1*d2*d3+d1*(x2*x3+y2*y3+z2*z3)
			      +d2*(x3*x1+y3*y1+z3*z1)
			      +d3*(x1*x2+y1*y2+z1*z2));
#ifdef	SI_P
	    p+=(f*g);
#endif
#ifdef	SI_G
	    gx-=(si[n].nx*g); gy-=(si[n].ny*g); gz-=(si[n].nz*g);
#endif
	}
#ifdef	SI_P
	*P=p/2.0;
#endif
#ifdef	SI_G
	*Gx=gx; *Gy=gy; *Gz=gz;
#endif
}

#define EndSI(si)	free(si)
