
#include <stdio.h>
#include "facet.h"

extern void	Error(char *msg),
		SurfaceIntegral(double x0,double y0,double z0,
				double x1,double y1,double z1,
				double x2,double y2,double z2,
				double x3,double y3,double z3,
				double *P,double *X,double *Y,double *Z);

#define LEN	1024

int	main(int argc,char **argv)
{
	size_t	N,n,m,l;
	Facet	*facet;
	double	V,x,y,z,v,P,X,Y,Z;
	FILE	*file;
	char	text[LEN];

	if (argc!=2 && argc!=3) Error("usage : stl_pxyz STL {point}");

	ReadFacet(argv[1],&N,&facet);

#define X1(n)	facet[n].x1
#define Y1(n)	facet[n].y1
#define Z1(n)	facet[n].z1
#define X2(n)	facet[n].x2
#define Y2(n)	facet[n].y2
#define Z2(n)	facet[n].z2
#define X3(n)	facet[n].x3
#define Y3(n)	facet[n].y3
#define Z3(n)	facet[n].z3

	V=x=y=z=0.0;
	for (n=0; n<N; n++) {
	    V+=(v=X1(n)*Y2(n)*Z3(n)+X2(n)*Y3(n)*Z1(n)+X3(n)*Y1(n)*Z2(n)
		 -X3(n)*Y2(n)*Z1(n)-X1(n)*Y3(n)*Z2(n)-X2(n)*Y1(n)*Z3(n));
	    x+=(v*(X1(n)+X2(n)+X3(n)));
	    y+=(v*(Y1(n)+Y2(n)+Y3(n)));
	    z+=(v*(Z1(n)+Z2(n)+Z3(n)));
	}
	v=V*4.0; (void)printf("%le\t%le\t%le\t%le\n",V/6.0,x/v,y/v,z/v);

	if (argc==2)
	    for (m=0; m<N; m++) {
		x=(X1(m)+X2(m)+X3(m))/3.0;
		y=(Y1(m)+Y2(m)+Y3(m))/3.0;
		z=(Z1(m)+Z2(m)+Z3(m))/3.0; P=X=Y=Z=0.0;
		for (l=0; l<N; l++)
		    SurfaceIntegral(x,y,z,X1(l),Y1(l),Z1(l),
					  X2(l),Y2(l),Z2(l),
					  X3(l),Y3(l),Z3(l),&P,&X,&Y,&Z);

		(void)printf("%le\t%le\t%le\t"
			     "%le\t%le\t%le\t"
			     "%le\t%le\t%le\t%le\t%le\t%le\t%le\n",
			     X1(m),Y1(m),Z1(m),
			     X2(m),Y2(m),Z2(m),
			     X3(m),Y3(m),Z3(m),P,X,Y,Z);
	    }
	else {
	    if (argv[2][0]=='-')
		file=stdin;
	    else
		if ((file=fopen(argv[2],"r"))==NULL)
		    Error("point file not found.");

	    while (fgets(text,LEN,file)!=NULL &&
		   sscanf(text,"%lf %lf %lf",&x,&y,&z)==3) {
		P=X=Y=Z=0.0;
		for (n=0; n<N; n++)
		    SurfaceIntegral(x,y,z,X1(n),Y1(n),Z1(n),
					  X2(n),Y2(n),Z2(n),
					  X3(n),Y3(n),Z3(n),&P,&X,&Y,&Z);

		(void)printf("%le\t%le\t%le\t%le\t%le\t%le\t%le\n",
			     x,y,z,P,X,Y,Z);
	    }
	    if (file!=stdin) (void)fclose(file);
	}
	return 0;
}
