
#include <stdio.h>

#define SI_P
#define SI_G
#include "si.h"

#include "mt.h"

static size_t	M,N,O;
static Facet	*facet;
static SI	*si;
static double	*X0,*Y0,*Z0,*P,*X,*Y,*Z;

static FUNCTION_TYPE	ExecuteSI(void *arg)
{
	size_t	o;

	for (o=(size_t)arg; o<O; o+=M)
	    CalculateSI(N,facet,si,X0[o],Y0[o],Z0[o],P+o,X+o,Y+o,Z+o);

	return RETURN_VALUE;
}

static void	AllocateMemory()
{
	if ((P=(double *)malloc(sizeof(double)*O))==NULL ||
	    (X=(double *)malloc(sizeof(double)*O))==NULL ||
	    (Y=(double *)malloc(sizeof(double)*O))==NULL ||
	    (Z=(double *)malloc(sizeof(double)*O))==NULL)
	    Error("memory allocation error for P, X, Y and Z.");
}

static THREAD_TYPE	*T;

static void	MultiThreading()
{
	size_t	m;

	for (m=1; m<M; m++)
	    if (INIT_THREAD(T[m],ExecuteSI,(size_t)m))
		Error("multi-threading initialization error.");

	(void)ExecuteSI(0);

	for (m=M-1; m>=1; m--)
	    if (TERM_THREAD(T[m])) Error("multi-threading termination error.");
}

#define LEN	1024

int	main(int argc,char **argv)
{
	size_t	n,o;
	double	V,x,y,z,v;
	FILE	*file;
	char	text[LEN];

	if (argc!=2 && argc!=3) Error("usage : stl_pxyz STL {point}");

	M=(size_t)BeginMT(&T);

	ReadFacet(argv[1],&N,&facet);

	si=BeginSI(N,facet);

	V=x=y=z=0.0;
	for (n=0; n<N; n++) {
	    V+=(v=X1(n)*Y2(n)*Z3(n)+X2(n)*Y3(n)*Z1(n)+X3(n)*Y1(n)*Z2(n)
		 -X3(n)*Y2(n)*Z1(n)-X1(n)*Y3(n)*Z2(n)-X2(n)*Y1(n)*Z3(n));
	    x+=(v*(X1(n)+X2(n)+X3(n)));
	    y+=(v*(Y1(n)+Y2(n)+Y3(n)));
	    z+=(v*(Z1(n)+Z2(n)+Z3(n)));
	}
	v=V*4.0; (void)printf("%le\t%le\t%le\t%le\n",V/6.0,x/v,y/v,z/v);

	if (argc==2) {
	    O=N; AllocateMemory(); X0=X; Y0=Y; Z0=Z;
	    for (n=0; n<N; n++) {
		X0[n]=(X1(n)+X2(n)+X3(n))/3.0;
		Y0[n]=(Y1(n)+Y2(n)+Y3(n))/3.0;
		Z0[n]=(Z1(n)+Z2(n)+Z3(n))/3.0;
	    }
	    MultiThreading();
	    for (n=0; n<N; n++)
		(void)printf("%le\t%le\t%le\t"
			     "%le\t%le\t%le\t"
			     "%le\t%le\t%le\t%le\t%le\t%le\t%le\n",
			     X1(n),Y1(n),Z1(n),
			     X2(n),Y2(n),Z2(n),
			     X3(n),Y3(n),Z3(n),P[n],X[n],Y[n],Z[n]);
	}
	else {
	    if (argv[2][0]=='-')
		file=stdin;
	    else
		if ((file=fopen(argv[2],"r"))==NULL)
		    Error("point file not found.");

	    O=0; X0=Y0=Z0=NULL;
	    while (fgets(text,LEN,file)!=NULL &&
		   sscanf(text,"%lf %lf %lf",&x,&y,&z)==3) {
		o=O++;
		if ((X0=(double *)realloc(X0,sizeof(double)*O))==NULL ||
		    (Y0=(double *)realloc(Y0,sizeof(double)*O))==NULL ||
		    (Z0=(double *)realloc(Z0,sizeof(double)*O))==NULL)
		    Error("memory re-allocation error for X0, Y0 and Z0.");

		X0[o]=x; Y0[o]=y; Z0[o]=z;
	    }
	    if (file!=stdin) (void)fclose(file);

	    if (O!=0) {
		AllocateMemory(); MultiThreading();
		for (o=0; o<O; o++)
		    (void)printf("%le\t%le\t%le\t%le\t%le\t%le\t%le\n",
				 X0[o],Y0[o],Z0[o],P[o],X[o],Y[o],Z[o]);
	    }
	}
	EndSI(si); EndMT(T); return 0;
}
